from flask import g, has_request_context, request, current_app
from frasco.ext import *
from frasco.users import is_user_logged_in, current_user
from frasco.models import delayed_tx_calls
from frasco.ctx import ContextStack
from frasco.assets import expose_package
from itsdangerous import URLSafeTimedSerializer
from redis import Redis
import hashlib
import logging
import subprocess
import sys
import pickle
import uuid
import click


suppress_push_events = ContextStack(False, default_item=True, ignore_nested=True)
testing_push_events = ContextStack(None, list, ignore_nested=True)
dont_skip_self_push_events = ContextStack(False, default_item=True, ignore_nested=True)
logger = logging.getLogger('frasco.push')


class FrascoPushState(ExtensionState):
    def __init__(self, *args, **kwargs):
        super(FrascoPushState, self).__init__(*args, **kwargs)
        self.current_user_loader = default_current_user_loader


class FrascoPush(Extension):
    name = 'frasco_push'
    state_class = FrascoPushState
    defaults = {"redis_url": None,
                "server_url": None,
                "server_port": 8888,
                "server_secured": False,
                "channel": "socketio",
                "secret": None,
                "prefix_event_with_room": True,
                "default_current_user_loader": True,
                "testing_ignore_redis_publish": True}

    def _init_app(self, app, state):
        expose_package(app, "frasco_push", __name__)

        if state.options['secret'] is None:
            state.options["secret"] = app.config['SECRET_KEY']

        if not state.options['redis_url'] and has_extension('frasco_redis', app):
            state.options['redis_url'] = app.extensions.frasco_redis.options['url']

        state.server_cli = ["python", "-m", "frasco.push.server",
            "--channel", state.options["channel"],
            "--redis", state.options["redis_url"],
            "--port", str(state.options["server_port"])]
        if state.options['secret']:
            state.server_cli.extend(["--secret", state.options["secret"]])
        if app.debug or app.testing:
            state.server_cli.append("--debug")

        if state.options["server_url"]:
            state.server_url = state.options['server_url']
        else:
            server_name = app.config.get('SERVER_NAME') or 'localhost'
            state.server_url = "%s://%s:%s" % (
                "https" if state.options['server_secured'] else "http",
                server_name.split(':')[0], state.options['server_port'])

        state.token_serializer = URLSafeTimedSerializer(state.options['secret'])
        state.redis = Redis.from_url(state.options["redis_url"])
        state.host_id = uuid.uuid4().hex

        @app.cli.command('push-server')
        @click.option('--access-logs', is_flag=True)
        @click.option('--debug', is_flag=True)
        def cli_server(access_logs=False, debug=False):
            """Start the push server"""
            args = list(state.server_cli)
            if access_logs:
                args.append("--access-logs")
            if debug:
                args.append("--debug")
            process = subprocess.Popen(args)
            process.wait()
            sys.exit(process.returncode)

        @app.cli.command('push-server-cli')
        def cli_print_cmd():
            """Print the command line to start the push server independently from the app"""
            click.echo(" ".join(state.server_cli))

        @app.before_request
        def before_request():
            if state.options['secret']:
                user_id, user_info, allowed_rooms = state.current_user_loader()
                g.socketio_token = create_push_token(user_info, get_user_room_name(user_id), allowed_rooms)

    @ext_stateful_method
    def current_user_loader(self, state, func):
        state.current_user_loader = func
        return func


def default_current_user_loader():
    if not is_user_logged_in():
        return None, {"guest": True}, None

    allowed_rooms = None
    if hasattr(current_user, 'get_allowed_push_rooms'):
        allowed_rooms = current_user.get_allowed_push_rooms()

    info = {"guest": False}
    info['username'] = getattr(current_user, 'username', current_user.email)
    if has_extension('frasco_users_avatar'):
        info['avatar_url'] = current_user.avatar_url

    return current_user.get_id(), info, allowed_rooms


def create_push_token(user_info=None, user_room=None, allowed_rooms=None):
    return get_extension_state('frasco_push').token_serializer.dumps([user_info, user_room, allowed_rooms])


@delayed_tx_calls.proxy
def emit_push_event(event, data=None, skip_self=None, room=None, namespace=None, prefix_event_with_room=True):
    if suppress_push_events.top:
        return
    state = get_extension_state('frasco_push')
    if current_app.testing and testing_push_events.top is not None:
        testing_push_events.top.append((event, data, skip_self, room, namespace))
        if state.options['testing_ignore_redis_publish']:
            return
    if state.options['prefix_event_with_room'] and prefix_event_with_room and room:
        event = "%s:%s" % (room, event)
    if skip_self is None:
        skip_self = not dont_skip_self_push_events.top
    skip_sid = None
    if skip_self and has_request_context() and 'x-socketio-sid' in request.headers:
        skip_sid = request.headers['x-socketio-sid']
    logger.debug("Push event '%s' to {namespace=%s, room=%s, skip_sid=%s}: %s" % (event, namespace, room, skip_sid, data))
    # we are publishing the events ourselves rather than using socketio.RedisManager because
    # importing socketio while using flask debugger causes an error due to signals
    return state.redis.publish(state.options['channel'], format_emit_data(event, data, namespace, room, skip_sid))


def emit_user_push_event(user_id, event, data=None, **kwargs):
    return emit_push_event(event, data, room=get_user_room_name(user_id), prefix_event_with_room=False, **kwargs)


def get_user_room_name(user_id):
    state = get_extension_state('frasco_push')
    if not state.options['secret']:
        raise Exception('A secret must be set to use emit_direct()')
    return hashlib.sha1((str(user_id) + state.options['secret']).encode('utf-8')).hexdigest()


def format_emit_data(event, data, namespace=None, room=None, skip_sid=None):
    # See https://github.com/miguelgrinberg/python-socketio/blob/master/socketio/pubsub_manager.py#L65
    return pickle.dumps({
        'method': 'emit',
        'event': event,
        'data': data,
        'namespace': namespace or '/',
        'room': room,
        'skip_sid': skip_sid,
        'callback': None,
        'host_id': get_extension_state('frasco_push').host_id
    })




import logging
import re


BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
#The background is set with 40 plus the number of the color, and the foreground with 30
#These are the sequences need to get colored ouput
RESET_COLOR_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"


def remove_all_handlers_from_logger(logger):
    for handler in logger.handlers:
        logger.removeHandler(handler)
    return logger


class ColorizingStreamHandler(logging.StreamHandler):
    def format(self, record):
        return self.colorize(super(ColorizingStreamHandler, self).format(record), record)

    def colorize(self, message, record):
        color = None
        if record.levelname == 'ERROR':
            color = RED
        if record.levelname == 'WARNING':
            color = YELLOW
        elif record.name.startswith('sqlalchemy.engine'):
            if re.match(r"(SELECT|WITH)", record.msg):
                color = BLUE
            elif re.match(r"(INSERT|UPDATE|DELETE)", record.msg):
                color = MAGENTA
            elif re.match(r"(BEGIN|COMMIT|ROLLBACK)", record.msg):
                color = GREEN
        if color:
            message = "\n".join([COLOR_SEQ % (30 + color) + line + RESET_COLOR_SEQ for line in message.split("\n")])
        return message


class ExcludeWerkzeugLogFilter(logging.Filter):
    def filter(self, record):
        return record.name != 'werkzeug'



from flask import has_request_context, request, current_app
from frasco.models import db
from frasco.ext import get_extension_state
from frasco.utils import cached_property
from sqlalchemy.ext.declarative import declared_attr
import stripe
from . import signals
import datetime
import time
import logging


logger = logging.getLogger('frasco.billing.stripe')


class StripeModelMixin(object):
    __stripe_auto_create_customer__ = True
    __stripe_email_property__ = 'email'
    __stripe_customer_expand__ = None

    stripe_customer_id = db.Column(db.String, index=True)
    has_stripe_payment_method = db.Column(db.Boolean, default=False)

    @classmethod
    def query_by_stripe_customer(cls, customer_id):
        return cls.query.filter(cls.stripe_customer_id == customer_id)

    @classmethod
    def create_stripe_payment_method(cls, type, **kwargs):
        return stripe.PaymentMethod.create(type=type, **kwargs)

    @cached_property
    def stripe_customer(self):
        try:
            return stripe.Customer.retrieve(self.stripe_customer_id, expand=self.__stripe_customer_expand__)
        except stripe.error.InvalidRequestError:
            if self.__stripe_auto_create_customer__:
                return self.create_stripe_customer()
            return

    @cached_property
    def default_stripe_invoice_payment_method(self):
        if self.stripe_customer.invoice_settings.default_payment_method:
            return stripe.PaymentMethod.retrieve(self.stripe_customer.invoice_settings.default_payment_method)

    def create_stripe_customer(self, **kwargs):
        kwargs.setdefault('email', getattr(self, self.__stripe_email_property__, None))
        kwargs.setdefault('expand', self.__stripe_customer_expand__)
        if kwargs.get('default_payment_method'):
            kwargs['payment_method'] = kwargs.pop('default_payment_method')
            kwargs.setdefault('invoice_settings', {})['default_payment_method'] = kwargs['payment_method']
        customer = stripe.Customer.create(**kwargs)
        self._update_stripe_customer(customer)
        logger.info('Customer %s for model #%s created' % (customer.id, self.id))
        return customer

    def update_stripe_customer(self, **kwargs):
        customer = self.stripe_customer
        for k, v in kwargs.items():
            setattr(customer, k, v)
        customer.save()
        self._update_stripe_customer(customer)
        logger.info('Customer %s updated' % self.stripe_customer_id)

    def delete_stripe_customer(self):
        if self.stripe_customer_id:
            try:
                self.stripe_customer.delete()
                logger.info('Customer %s deleted' % self.stripe_customer_id)
            except stripe.error.InvalidRequestError as e:
                if 'No such customer' not in str(e):
                    logger.warning("Customer %s that was to be deleted didn't exist anymore in Stripe" % self.stripe_customer_id)
                    raise
        self._update_stripe_customer(False)

    def _update_stripe_customer(self, customer=None):
        if customer is None:
            customer = self.stripe_customer

        if customer:
            self.stripe_customer_id = customer.id
            self.has_stripe_payment_method = bool(customer.invoice_settings['default_payment_method'])
            self.__dict__['stripe_customer'] = customer
        else:
            self.stripe_customer_id = None
            self.has_stripe_payment_method = False
            self.__dict__.pop('stripe_customer', None)

    def add_stripe_payment_method(self, type, make_default=True, **kwargs):
        pm = self.create_stripe_payment_method(type, **kwargs)
        self.attach_stripe_payment_method(pm.id, make_default)
        return pm

    def attach_stripe_payment_method(self, payment_method_id, make_default=True):
        stripe.PaymentMethod.attach(payment_method_id, customer=self.stripe_customer_id)
        if make_default:
            self.update_stripe_customer(invoice_settings=dict(self.stripe_customer.invoice_settings, default_payment_method=payment_method_id))

    def start_stripe_payment_intent(self, amount, currency, **kwargs):
        return stripe.PaymentIntent.create(amount=amount, currency=currency, customer=self.stripe_customer_id, **kwargs)


class StripeSubscriptionModelMixin(StripeModelMixin):
    __stripe_subscription_expand__ = None
    stripe_subscription_id = db.Column(db.String, index=True)
    
    @declared_attr
    def plan_name(cls):
        return db.Column(db.String, index=True)
    
    @declared_attr        
    def plan_status(cls):
        return db.Column(db.String, default='trialing', index=True)
    
    @declared_attr        
    def plan_last_invoice_at(cls):
        return db.Column(db.DateTime)
    
    @declared_attr        
    def plan_last_invoice_amount(cls):
        return db.Column(db.Float)
    
    @declared_attr        
    def plan_current_period_start(cls):
        return db.Column(db.DateTime)
    
    @declared_attr        
    def plan_current_period_end(cls):
        return db.Column(db.DateTime)
    
    @declared_attr        
    def plan_next_charge_at(cls):
        return db.Column(db.DateTime)
    
    @declared_attr        
    def plan_cancel_at_period_end(cls):
        return db.Column(db.Boolean, default=False)
    
    @declared_attr        
    def plan_cancelled_at(cls):
        return db.Column(db.DateTime)
    
    @declared_attr        
    def plan_has_invoice_items(cls):
        return db.Column(db.Boolean, index=True, default=False)
    
    @declared_attr        
    def plan_last_invoice_item_added_at(cls):
        return db.Column(db.DateTime, index=True)

    @classmethod
    def query_by_stripe_subscription(cls, subscription_id):
        return cls.query.filter(cls.stripe_subscription_id == subscription_id)

    @classmethod
    def query_has_stripe_invoice_items(cls, added_days_ago=None):
        q = cls.query.filter(cls.plan_has_invoice_items==True)
        if added_days_ago:
            q = q.filter(cls.plan_last_invoice_item_added_at<=datetime.date.today() - datetime.timedelta(days=added_days_ago))
        return q

    @cached_property
    def stripe_subscription(self):
        if not self.stripe_customer_id or not self.stripe_subscription_id:
            return
        expand = ['latest_invoice.payment_intent']
        if self.__stripe_subscription_expand__:
            expand.extend(self.__stripe_subscription_expand__)
        try:
            return self.stripe_customer.subscriptions.retrieve(self.stripe_subscription_id, expand=expand)
        except stripe.error.InvalidRequestError:
            return

    @property
    def stripe_subscription_item(self):
        sub = self.stripe_subscription
        if sub:
            return sub['items']['data'][0]

    def create_stripe_customer(self, plan=None, quantity=1, trial_end=None, coupon=None, **kwargs):
        state = get_extension_state('frasco_stripe')
        customer = super(StripeSubscriptionModelMixin, self).create_stripe_customer(**kwargs)

        if plan:
            self.subscribe_stripe_plan(plan, quantity, trial_end=trial_end, coupon=coupon)
        elif state.options['default_plan']:
            self.subscribe_stripe_plan(state.options['default_plan'], quantity, trial_end=trial_end, coupon=coupon)
                
        return customer

    def delete_stripe_customer(self):
        super(StripeSubscriptionModelMixin, self).delete_stripe_customer()
        if self.stripe_subscription_id:
            self._update_stripe_subscription(False)

    def subscribe_stripe_plan(self, plan=None, quantity=1, item_options=None, trial_end=None, **kwargs):
        state = get_extension_state('frasco_stripe')
        if not plan:
            plan = state.options['default_plan']
        if self.stripe_subscription_id and self.plan_name == plan:
            return

        item = dict(plan=plan, quantity=quantity)
        if item_options:
            item.update(item_options)

        kwargs.setdefault('expand', []).append('latest_invoice.payment_intent')
        if self.__stripe_subscription_expand__:
            kwargs['expand'].extend(self.__stripe_subscription_expand__)

        subscription = stripe.Subscription.create(customer=self.stripe_customer_id, items=[item],
            trial_end=_format_trial_end(trial_end), **kwargs)
        self._update_stripe_subscription(subscription)
        logger.info('Subscribed customer %s to plan %s' % (self.stripe_customer_id, plan))
        return subscription

    def update_stripe_subscription(self, **kwargs):
        kwargs.setdefault('expand', []).append('latest_invoice.payment_intent')
        if self.__stripe_subscription_expand__:
            kwargs['expand'].extend(self.__stripe_subscription_expand__)
        if 'trial_end' in kwargs:
            kwargs['trial_end'] = _format_trial_end(kwargs['trial_end'])
        subscription = stripe.Subscription.modify(self.stripe_subscription_id, **kwargs)
        self._update_stripe_subscription(subscription)
        logger.info('Subscription %s updated for %s' % (self.stripe_subscription_id, self.stripe_customer_id))

    def update_stripe_subscription_item(self, subscription_options=None, **kwargs):
        if not subscription_options:
            subscription_options = {}
        item = dict(id=self.stripe_subscription['items']['data'][0].id, **kwargs)
        self.update_stripe_subscription(items=[item], **subscription_options)

    def change_stripe_subscription_plan(self, plan, quantity=None, cancel_at_period_end=False, proration_behavior='create_prorations'):
        kwargs = {'plan': plan,
                  'quantity': quantity if quantity is not None else self.stripe_subscription_item['quantity']}
        self.update_stripe_subscription_item(subscription_options={
            'cancel_at_period_end': cancel_at_period_end,
            'proration_behavior': proration_behavior
        }, **kwargs)

    def cancel_stripe_subscription(self, at_period_end=True):
        if at_period_end:
            self.update_stripe_subscription(cancel_at_period_end=True)
        else:
            self.stripe_subscription.delete()
            self._update_stripe_subscription(False)
        logger.info('Subscription %s cancelled for %s' % (self.stripe_subscription_id, self.stripe_customer_id))

    def revert_stripe_subscription_period_end_cancellation(self):
        self.update_stripe_subscription(cancel_at_period_end=False)
        logger.info('Subscription %s reverted cancellation for %s' % (self.stripe_subscription_id, self.stripe_customer_id))

    def _update_stripe_customer(self, customer=None):
        super(StripeSubscriptionModelMixin, self)._update_stripe_customer(customer)
        if not self.stripe_customer:
            self._update_stripe_subscription(False)

    def _update_stripe_subscription(self, subscription=None):
        if subscription is None:
            if self.stripe_subscription_id:
                self.__dict__.pop('stripe_subscription', None)
                subscription = self.stripe_subscription
            elif self.stripe_customer.subscriptions.total_count > 0:
                subscription = self.stripe_customer.subscriptions.data[0]

        prev_plan = self.plan_name
        prev_status = self.plan_status

        if subscription:
            self.stripe_subscription_id = subscription.id
            self.__dict__['stripe_subscription'] = subscription
            self.plan_name = subscription.plan.id
            self.plan_status = subscription.status
            self.plan_current_period_start = datetime.datetime.fromtimestamp(subscription.current_period_start)
            self.plan_current_period_end = datetime.datetime.fromtimestamp(subscription.current_period_end)
            self.plan_cancel_at_period_end = subscription.cancel_at_period_end
            if self.plan_status == 'trialing':
                self.plan_next_charge_at = datetime.datetime.fromtimestamp(subscription.trial_end)
            elif subscription.current_period_end:
                self.plan_next_charge_at = self.plan_current_period_end
            else:
                self.plan_next_charge_at = None
            if self.plan_status == 'canceled' and prev_status != 'canceled':
                self.plan_cancelled_at = datetime.datetime.utcnow()
        else:
            if self.stripe_subscription_id:
                self.plan_cancelled_at = datetime.datetime.utcnow()
            self.stripe_subscription_id = None
            self.__dict__.pop('stripe_subscription', None)
            self.plan_name = None
            self.plan_status = 'canceled'
            self.plan_current_period_start = None
            self.plan_current_period_end = None
            self.plan_next_charge_at = None
            self.plan_cancel_at_period_end = None

        signals.model_subscription_updated.send(self, prev_plan=prev_plan, prev_status=prev_status)

    def is_stripe_subscription_payment_method_missing(self):
        return self.plan_status in ('incomplete', 'past_due') and self.stripe_subscription.latest_invoice.payment_intent.status == 'requires_payment_method'

    def does_stripe_subscription_payment_method_requires_action(self):
        return self.plan_status in ('incomplete', 'past_due') and self.stripe_subscription.latest_invoice.payment_intent.status == 'requires_action'

    @property
    def stripe_subscription_payment_intent_secret(self):
        return self.stripe_subscription.latest_invoice.payment_intent.client_secret

    def reattempt_stripe_subscription_last_invoice_payment(self):
        return stripe.Invoice.pay(self.stripe_subscription.latest_invoice, expand=['payment_intent'])

    def update_last_stripe_subscription_invoice(self, invoice):
        self.plan_last_invoice_at = datetime.datetime.fromtimestamp(invoice.created)
        self.plan_last_invoice_amount = invoice.total / 100
        self._update_stripe_subscription()
        if invoice.next_payment_attempt:
            self.plan_next_charge_at = datetime.datetime.fromtimestamp(invoice.next_payment_attempt)
        signals.model_last_invoice_updated.send(self)

    def add_item_to_next_stripe_subscription_invoice(self, amount, currency, description=None, **kwargs):
        if description:
            kwargs['description'] = description
        item = stripe.InvoiceItem.create(customer=self.stripe_customer_id, amount=amount, currency=currency, **kwargs)
        if not self.plan_has_invoice_items:
            self.plan_has_invoice_items = True
            self.plan_last_invoice_item_added_at = datetime.datetime.utcnow()
        return item

    def charge_now_items_from_upcoming_stripe_subscription_invoice(self, charge=True, **kwargs):
        invoice = stripe.Invoice.create(customer=self.stripe_customer_id, **kwargs)
        self.plan_has_invoice_items = False
        if charge:
            return stripe.Invoice.pay(invoice.id)
        return stripe.Invoice.finalize_invoice(invoice.id)


class StripeBillingDetailsModelMixin(object):
    __stripe_has_billing_fields__ = True
    __stripe_email_property__ = 'billing_email'

    @declared_attr
    def billing_name(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_email(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_phone(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_address_line1(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_address_line2(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_address_city(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_address_state(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_address_zip(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_address_country(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_country(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_type(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_brand(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_exp_month(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_exp_year(cls):
        return db.Column(db.String)

    @declared_attr
    def billing_last4(cls):
        return db.Column(db.String)

    def create_stripe_customer(self, **kwargs):
        kwargs.update(self._get_billing_details_stripe_params())
        return super(StripeBillingDetailsModelMixin, self).create_stripe_customer(**kwargs)

    def _update_stripe_customer(self, customer=None):
        super(StripeBillingDetailsModelMixin, self)._update_stripe_customer(customer)
        if self.stripe_customer:
            self.update_payment_method_details_from_stripe()
        
    def update_from_stripe_billing_details(self):
        self.billing_name = self.stripe_customer.name
        self.billing_email = self.stripe_customer.email
        self.billing_phone = self.stripe_customer.phone
        addr = self.stripe_customer.address
        self.billing_address_line1 = addr['line1']
        self.billing_address_line2 = addr['line2']
        self.billing_address_zip = addr['postal_code']
        self.billing_address_city = addr['city']
        self.billing_address_state = addr['state']
        self.billing_address_country = addr['country']

    def update_stripe_billing_details(self, **kwargs):
        kwargs.update(self._get_billing_details_stripe_params())
        stripe.Customer.modify(self.stripe_customer_id, **kwargs)

    def _get_billing_details_stripe_params(self):
        o = {
            'name': self.billing_name,
            'email': self.billing_email,
            'phone': self.billing_phone
        }
        if self.billing_address_line1:
            o['address'] = {
                'line1': self.billing_address_line1,
                'line2': self.billing_address_line2,
                'postal_code': self.billing_address_zip,
                'city': self.billing_address_city,
                'state': self.billing_address_state,
                'country': self.billing_address_country
            }
        return o

    def update_payment_method_details_from_stripe(self, payment_method=None):
        if not payment_method:
            if not self.stripe_customer.invoice_settings.default_payment_method:
                return
            payment_method = stripe.PaymentMethod.retrieve(self.stripe_customer.invoice_settings.default_payment_method)

        self.billing_type = payment_method.type
        if payment_method.type == 'card':
            self.billing_last4 = payment_method.card['last4']
            self.billing_brand = payment_method.card['brand']
            self.billing_exp_month = payment_method.card['exp_month']
            self.billing_exp_year = payment_method.card['exp_year']
            self.billing_country = payment_method.card['country']
        elif payment_method.type == 'sepa_debit':
            self.billing_last4 = payment_method.sepa_debit['last4']
            self.billing_country = payment_method.sepa_debit['country']


class StripeEUVATModelMixin(object):
    __stripe_has_eu_vat__ = True
    __stripe_eu_auto_vat_country__ = True
    __stripe_eu_auto_vat_rate__ = True
    __stripe_eu_vat_use_address_country__ = False
    __stripe_eu_vat_exempt_reverse_charge__ = 'exempt'

    def create_stripe_customer(self, **kwargs):
        eu_vat_number = kwargs.pop('eu_vat_number', self.eu_vat_number)
        if eu_vat_number:
            kwargs['tax_id_data'] = [{'type': 'eu_vat', 'value': eu_vat_number}]
            kwargs['tax_exempt'] = 'none' if self.should_charge_eu_vat() else self.__stripe_eu_vat_exempt_reverse_charge__
        return super(StripeEUVATModelMixin, self).create_stripe_customer(**kwargs)

    def _update_stripe_customer(self, customer=None):
        super(StripeEUVATModelMixin, self)._update_stripe_customer(customer)
        if self.stripe_customer and self.__stripe_eu_auto_vat_country__:
            country = self.billing_address_country if self.__stripe_eu_vat_use_address_country__ else self.billing_country
            if country:
                self.eu_vat_country = country

    def update_from_stripe_eu_vat_number(self, customer=None):
        if not customer:
            customer = self.stripe_customer
        for tax_id in customer.tax_ids.data:
            if tax_id['type'] == 'eu_vat':
                self.eu_vat_number = tax_id['value']
                return

    def update_stripe_customer_eu_vat_number(self, eu_vat_number=None, tax_exempt=False, clear_existing=True):
        if clear_existing:
            self.remove_all_stripe_customer_tax_ids()
        if not eu_vat_number:
            eu_vat_number = self.eu_vat_number
            if eu_vat_number:
                tax_exempt = not self.should_charge_eu_vat()
        if eu_vat_number:
            stripe.Customer.create_tax_id(self.stripe_customer_id, type='eu_vat', value=eu_vat_number or self.eu_vat_number)
            stripe.Customer.modify(self.stripe_customer_id, tax_exempt=self.__stripe_eu_vat_exempt_reverse_charge__ if tax_exempt else 'none')
        else:
            stripe.Customer.modify(self.stripe_customer_id, tax_exempt='none')

    def remove_all_stripe_customer_tax_ids(self):
        for tax_id in stripe.Customer.list_tax_ids(self.stripe_customer_id):
            stripe.Customer.delete_tax_id(self.stripe_customer_id, tax_id.id)

    @property
    def stripe_default_tax_rates(self):
        if self.eu_vat_country:
            return [self._get_or_create_stripe_tax_rate(self.eu_vat_country, self.eu_vat_rate).id]

    def subscribe_stripe_plan(self, *args, **kwargs):
        if 'default_tax_rates' not in kwargs and self.__stripe_eu_auto_vat_rate__:
            kwargs['default_tax_rates'] = self.stripe_default_tax_rates
        return super(StripeEUVATModelMixin, self).subscribe_stripe_plan(*args, **kwargs)

    def update_stripe_subscription_tax_rates(self):
        stripe.Subscription.modify(self.stripe_subscription_id, default_tax_rates=self.stripe_default_tax_rates or '')

    def charge_now_items_from_upcoming_stripe_subscription_invoice(self, *args, **kwargs):
        if 'default_tax_rates' not in kwargs and self.__stripe_eu_auto_vat_rate__:
            kwargs['default_tax_rates'] = self.stripe_default_tax_rates
        return super(StripeEUVATModelMixin, self).charge_now_items_from_upcoming_stripe_subscription_invoice(*args, **kwargs)

    def _get_or_create_stripe_tax_rate(self, eu_country, tax_rate):
        for rate in stripe.TaxRate.list():
            if rate.jurisdiction == eu_country.upper():
                return rate
        return stripe.TaxRate.create(display_name="VAT", description="%s VAT" % eu_country.upper(),
            jurisdiction=eu_country.upper(), percentage=tax_rate, inclusive=False)


def _format_trial_end(trial_end=None):
    state = get_extension_state('frasco_stripe')
    if state.options['debug_trial_period'] and current_app.debug:
        if state.options['debug_trial_period'] == 'now':
            return 'now'
        else:
            trial_end = datetime.datetime.now() + \
                datetime.timedelta(days=state.options['debug_trial_period'])
    if isinstance(trial_end, datetime.datetime):
        if trial_end <= datetime.datetime.now():
            return 'now'
        return int(time.mktime(trial_end.timetuple()))
    return trial_end



from flask import current_app, url_for as flask_url_for, Markup, request
import functools


def url_for(*args, **kwargs):
    if kwargs.get('_external') and current_app.config.get('FORCE_URL_SCHEME'):
        kwargs.setdefault('_scheme', current_app.config['FORCE_URL_SCHEME'])
    return flask_url_for(*args, **kwargs)


def url_for_static(filename, **kwargs):
    """Shortcut function for url_for('static', filename=filename)
    """
    return url_for('static', filename=filename, **kwargs)


def url_for_same(**overrides):
    return url_for(request.endpoint, **dict(dict(request.args,
        **request.view_args), **overrides))


def get_remote_addr():
    if current_app.debug and "__remoteaddr" in request.values:
        return request.values["__remoteaddr"]
    return request.remote_addr


def wrap_in_markup(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return Markup(f(*args, **kwargs))
    return wrapper


def inject_app_config(app, config, prefix=None):
    for k, v in config.items():
        if prefix:
            k = "%s%s" % (prefix, k)
        app.config[k.upper()] = v



import socketio
from socketio.exceptions import ConnectionRefusedError
import os
import urllib.parse
import uuid
import json
from itsdangerous import URLSafeTimedSerializer, BadSignature
from eventlet import wsgi
import eventlet
import logging


eventlet.sleep()
eventlet.monkey_patch()


logger = logging.getLogger('frasco.push.server')


class PresenceEnabledRedisManager(socketio.RedisManager):
    def __init__(self, *args, **kwargs):
        self.presence_session_id = kwargs.pop('presence_session_id', None) or ''
        self.presence_key_prefix = "presence%s:" % self.presence_session_id
        super(PresenceEnabledRedisManager, self).__init__(*args, **kwargs)

    def _emit_joined(self, room, sid, info):
        self.server.emit('%s:joined' % room, {"sid": sid, "info": info}, room=room, skip_sid=sid)

    def enter_room(self, sid, namespace, room, skip_presence=False):
        super(PresenceEnabledRedisManager, self).enter_room(sid, namespace, room)
        if room and room != sid and not skip_presence:
            self.redis.sadd("%s%s:%s" % (self.presence_key_prefix, namespace, room), sid)
            self._emit_joined(room, sid, self.get_member_info(sid, namespace))

    def leave_room(self, sid, namespace, room):
        super(PresenceEnabledRedisManager, self).leave_room(sid, namespace, room)
        if room and room != sid:
            if self.redis.srem("%s%s:%s" % (self.presence_key_prefix, namespace, room), sid):
                self.server.emit('%s:left' % room, sid, room=room, skip_sid=sid)

    def get_room_members(self, namespace, room):
        return [sid.decode() for sid in self.redis.smembers("%s%s:%s" % (self.presence_key_prefix, namespace, room))]

    def set_member_info(self, sid, namespace, info):
        self.redis.set("%s%s@%s" % (self.presence_key_prefix, namespace, sid), json.dumps(info))
        for room in self.get_rooms(sid, namespace):
            if room != sid:
                self._emit_joined(room, sid, info)

    def get_member_info(self, sid, namespace):
        data = self.redis.get("%s%s@%s" % (self.presence_key_prefix, namespace, sid))
        if data:
            try:
                return json.loads(data)
            except:
                pass
        return {}

    def disconnect(self, sid, namespace):
        super(PresenceEnabledRedisManager, self).disconnect(sid, namespace)
        self.redis.delete("%s%s@%s" % (self.presence_key_prefix, namespace, sid))

    def cleanup_presence_keys(self):
        keys = self.redis.keys('%s*' % self.presence_key_prefix)
        pipe = self.redis.pipeline()
        for key in keys:
            pipe.delete(key)
        pipe.execute()


class PresenceEnabledServer(socketio.Server):
    def enter_room(self, sid, room, namespace=None, skip_presence=False):
        namespace = namespace or '/'
        self.logger.debug('%s is entering room %s [%s]', sid, room, namespace)
        self.manager.enter_room(sid, namespace, room, skip_presence)


def create_app(redis_url='redis://', channel='socketio', secret=None, presence_session_id=None, token_max_age=None, debug=False):
    mgr = PresenceEnabledRedisManager(redis_url, channel=channel, presence_session_id=presence_session_id)
    sio = PresenceEnabledServer(client_manager=mgr, async_mode='eventlet', logger=debug, cors_allowed_origins='*') # client must be identified via url token so cors to * is not a big risk
    token_serializer = URLSafeTimedSerializer(secret)
    default_ns = '/'

    @sio.on('connect')
    def connect(sid, env):
        if not secret:
            raise ConnectionRefusedError('no secret defined')

        qs = urllib.parse.parse_qs(env['QUERY_STRING'])
        if not 'token' in qs:
            raise ConnectionRefusedError('missing token')

        try:
            token_data = token_serializer.loads(qs['token'][0], max_age=token_max_age)
        except BadSignature:
            logger.debug('Client provided an invalid token')
            raise ConnectionRefusedError('invalid token')

        if len(token_data) == 3:
            user_info, user_room, allowed_rooms = token_data
        else:
            # old format
            user_info, allowed_rooms = token_data
            user_room = None

        env['allowed_rooms'] = allowed_rooms
        if user_info:
            mgr.set_member_info(sid, default_ns, user_info)
        if user_room:
            sio.enter_room(sid, user_room, skip_presence=True)

        logger.debug('New client connection: %s ; %s' % (sid, user_info))
        return True

    @sio.on('members')
    def get_room_members(sid, data):
        if not data.get('room') or data['room'] not in mgr.get_rooms(sid, default_ns):
            return []
        return {sid: mgr.get_member_info(sid, default_ns) for sid in mgr.get_room_members(default_ns, data['room'])}

    @sio.on('join')
    def join(sid, data):
        if sio.environ[sid].get('allowed_rooms') is not None and data['room'] not in sio.environ[sid]['allowed_rooms']:
            logger.debug('Client %s is not allowed to join room %s' % (sid, data['room']))
            return False
        sio.enter_room(sid, data['room'])
        logger.debug('Client %s has joined room %s' % (sid, data['room']))
        return get_room_members(sid, data)

    @sio.on('broadcast')
    def room_broadcast(sid, data):
        logger.debug('Client %s broadcasting %s to room %s' % (sid, data['event'], data['room']))
        sio.emit("%s:%s" % (data['room'], data['event']), data.get('data'), room=data['room'], skip_sid=sid)

    @sio.on('leave')
    def leave(sid, data):
        sio.leave_room(sid, data['room'])
        logger.debug('Client %s has left room %s' % (sid, data['room']))

    @sio.on('set')
    def set(sid, data):
        mgr.set_member_info(sid, default_ns, data)
        logger.debug('Client %s has updated its user info: %s' % (sid, data))

    @sio.on('get')
    def get(sid, data):
        return mgr.get_member_info(data['sid'], default_ns)

    return socketio.WSGIApp(sio)


_wsgi_app = None
def wsgi_app(environ, start_response):
    global _wsgi_app
    if not _wsgi_app:
        _wsgi_app = create_app(os.environ.get('SIO_REDIS_URL', 'redis://'),
            os.environ.get('SIO_CHANNEL', 'socketio'), os.environ.get('SIO_SECRET'),
            os.environ.get('SIO_PRESENCE_SESSION_ID'), os.environ.get('SIO_DEBUG', False))
    return _wsgi_app(environ, start_response)


def run_server(port=8888, access_logs=False, reuse_addr=False, presence_session_id=None, **kwargs):
    logger.addHandler(logging.StreamHandler())
    debug = kwargs.get('debug', False)
    if debug:
        logger.setLevel(logging.DEBUG)
        logger.debug('Push server running in DEBUG')

    cleanup_presence_keys = False
    if not presence_session_id:
        presence_session_id = str(uuid.uuid4()).split('-')[0]
        logger.info('Generated random presence session id: %s' % presence_session_id)
        cleanup_presence_keys = True

    kwargs['presence_session_id'] = presence_session_id
    app = create_app(**kwargs)
    socket = eventlet.listen(('', port), reuse_addr=reuse_addr)
    wsgi.server(socket, app, debug=debug, log_output=debug or access_logs)

    if cleanup_presence_keys:
        app.engineio_app.manager.cleanup_presence_keys()


if __name__ == '__main__':
    import argparse
    argparser = argparse.ArgumentParser(prog='frascopush',
        description='Start frasco.push.server')
    argparser.add_argument('-p', '--port', default=8888, type=int,
        help='Port number')
    argparser.add_argument('-r', '--redis-url', default=os.environ.get('SIO_REDIS_URL', 'redis://'), type=str,
        help='Redis URL')
    argparser.add_argument('-c', '--channel', default=os.environ.get('SIO_CHANNEL', 'socketio'), type=str,
        help='Redis channel')
    argparser.add_argument('-s', '--secret', default=os.environ.get('SIO_SECRET'), type=str,
        help='Secret')
    argparser.add_argument('--presence-session-id', default=os.environ.get('SIO_PRESENCE_SESSION_ID'), type=str,
        help='Presence session id (when using multiple servers, use the same presence session id on all instances)')
    argparser.add_argument('--debug', action='store_true', help='Debug mode')
    argparser.add_argument('--access-logs', action='store_true', help='Show access logs in console')
    argparser.add_argument('--reuse-addr', action='store_true', help='Reuse address and port if already bound')
    args = argparser.parse_args()
    run_server(args.port, debug=args.debug, access_logs=args.access_logs,
        redis_url=args.redis_url, channel=args.channel, secret=args.secret,
        presence_session_id=args.presence_session_id, reuse_addr=args.reuse_addr)



from frasco.ext import *
from frasco.assets import expose_package, register_assets_builder
from frasco.utils import join_url_rule
from flask import render_template
import os
import json
import re
import htmlmin
import codecs


class FrascoAngular(Extension):
    name = "frasco_angular"
    defaults = {"static_dir": None, # defaults to app.static_folder
                "static_url_path": None, # defaults to app.static_url_path
                "angular_template": "angular_layout.html",
                "app_dir": "app",
                "services_module": "services",
                "services_name": "%s",
                "templates_file": None,
                "templates_module": "templatesCache",
                "disable_templates_cache": None, # app.debug
                "templates_matcher": r".*\.html$",
                "add_app_dir_in_babel_extract": True}

    def _init_app(self, app, state):
        require_extension('frasco_assets', app)
        expose_package(app, "frasco_angular", __name__)

        if not state.options["static_dir"]:
            state.options["static_dir"] = app.static_folder
        if not state.options["static_url_path"]:
            state.options["static_url_path"] = app.static_url_path

        if state.options['templates_file']:
            register_assets_builder(self.build_templates)

        if has_extension('frasco_babel', app) and state.options['add_app_dir_in_babel_extract']:
            app.extensions.frasco_babel.add_extract_dir(os.path.join(state.options['static_dir'], state.options['app_dir']),
                '.', ['frasco.angular.babel.AngularCompatExtension'], [('javascript:**.js', {})])

    @ext_stateful_method
    def add_route(self, state, endpoint, rule, decorators=None, **options):
        rules = rule if isinstance(rule, (list, tuple)) else [rule]
        def func(*args, **kwargs):
            return self.angular_view_response()
        if decorators:
            for decorator in reversed(decorators):
                func = decorator(func)
        for rule in rules:
            self.get_app().add_url_rule(rule, endpoint, func, **options)

    def angular_view_response(self):
        return render_template(get_extension_state('frasco_angular').options['angular_template'])

    @ext_stateful_method
    def register_service_builder(self, state, api_version, filename):
        def builder():
            module = ("/* This file is auto-generated by frasco-angular. DO NOT MODIFY. */\n'use strict';\n"
                    "\n(function() {\n\nvar services = angular.module('%s', ['frasco']);\n") % state.options["services_module"]

            for service in api_version.iter_services():
                endpoints = {}
                for rule, endpoint, func, options in service.iter_endpoints():
                    args = []
                    if hasattr(func, 'request_params'):
                        for p in reversed(func.request_params):
                            args.extend(p.names)
                    endpoints[endpoint] = [_convert_url_args(join_url_rule(service.url_prefix, rule)), args]
                module += ("\nservices.factory('%s', ['frascoServiceFactory', function(frascoServiceFactory) {\n"
                        "return frascoServiceFactory.make('%s', '%s', [], %s);\n}]);\n") % \
                            (state.options['services_name'] % service.name, service.name, api_version.url_prefix,
                            json.dumps(endpoints, indent=2))

            module += "\n})();"
            _write_file(os.path.join(state.options["static_dir"], state.options["app_dir"], filename), module)
        register_assets_builder(builder)

    @ext_stateful_method
    def build_templates(self, state):
        module = [("/* This file is auto-generated by frasco-angular. DO NOT MODIFY. */\n'use strict';\n"
                  "\nangular.module('%s', []).run(['$templateCache', function($templateCache) {") % state.options["templates_module"]]
        matcher = re.compile(state.options["templates_matcher"], re.I)
        done = set()

        def process_file(filename, path=None, content=None):
            if not path:
                pathname = filename
                path = os.path.dirname(filename)
                filename = os.path.basename(filename)
            else:
                pathname = os.path.join(path, filename)
            relname = state.options["static_url_path"] + "/" + os.path.relpath(path, state.options["static_dir"]) + "/" + filename
            if pathname not in done and matcher.match(relname):
                if not content:
                    with codecs.open(pathname, 'r', 'utf-8') as f:
                        content = f.read()
                module.append("  $templateCache.put('%s', %s);" % (relname, json.dumps(htmlmin.minify(content))))
                done.add(pathname)

        disable = state.options["disable_templates_cache"]
        if (disable is None and not self.get_app().debug) or disable is False:
            path = os.path.join(state.options["static_dir"], state.options['app_dir'])
            for path, dirnames, filenames in os.walk(os.path.join(state.options["static_dir"], state.options['app_dir'])):
                for filename in filenames:
                    process_file(filename, path)

        module = "\n".join(module) + "\n}]);"
        filename = os.path.join(state.options["static_dir"], state.options["app_dir"], state.options['templates_file'])
        _write_file(filename, module)


_url_arg_re = re.compile(r"<([a-z]+:)?([a-z0-9_]+)>")
def _convert_url_args(url):
    return _url_arg_re.sub(r":\2", url)


def _write_file(filename, source):
    if not os.path.exists(os.path.dirname(filename)):
        os.makedirs(os.path.dirname(filename))
    with codecs.open(filename, "w", "utf-8") as f:
        f.write(source)



from flask import json
from apispec import APISpec
from frasco.utils import join_url_rule
import re


__all__ = ('build_swagger_spec',)


def build_swagger_spec(api_version, with_security_scheme=False):
    spec = APISpec(title="API %s" % api_version.version,
                    version=api_version.version,
                    openapi_version="3.0.2",
                    basePath=api_version.url_prefix)

    if with_security_scheme:
        spec.components.security_scheme("api_key_bearer_token", {"type": "http", "scheme": "bearer"})
        spec.components.security_scheme("api_key_header", {"type": "apiKey", "in": "header", "name": "X-Api-Key"})
        spec.components.response("NotAuthenticatedError", {"description": "Authentification required"})
        spec.components.response("NotAuthorizedError", {"description": "Some permissions are missing to perform this request"})

    for service in api_version.iter_services():
        paths = {}
        tag = {"name": service.name}
        if hasattr(service, 'description'):
            tag["description"] = service.description
        spec.tag(tag)
        for rule, endpoint, func, options in service.iter_endpoints():
            rule = join_url_rule(service.url_prefix, rule)
            path = paths.setdefault(convert_url_args(rule), {})
            for method in options.get('methods', ['GET']):
                op = build_spec_operation(rule, service.name + '_' + endpoint, func, options, with_security_scheme)
                op['tags'] = [service.name]
                path[method.lower()] = op
        for path, operations in paths.items():
            spec.path(path=path, operations=operations)

    return spec


def build_spec_operation(rule, endpoint, func, options, with_security_scheme=False):
    responses = {
        "200": {"description": "Successful response"},
        "default": {"description": "Unexpected error"}
    }
    if with_security_scheme:
        responses["401"] = {"$ref": "#/components/responses/NotAuthenticatedError"}
        responses["403"] = {"$ref": "#/components/responses/NotAuthorizedError"}

    path_request_params = []
    query_request_params = []
    body_request_params = []
    file_request_params = []
    if hasattr(func, 'request_params'):
        url = convert_url_args(rule)
        method = options.get('methods', ['GET'])[0]
        for p in reversed(func.request_params):
            for pname in p.names:
                if p.location == 'files':
                    file_request_params.append((pname, p))
                elif ("{%s}" % pname) in url:
                    path_request_params.append((pname, p))
                elif method == 'GET':
                    query_request_params.append((pname, p))
                else:
                    body_request_params.append((pname, p))

    params = [build_spec_param(n, p, "path") for (n, p) in path_request_params] \
           + [build_spec_param(n, p) for (n, p) in query_request_params]

    schema = None
    if body_request_params:
        schema = {
            "type": "object",
            "required": [n for (n, p) in body_request_params if bool(p.required)],
            "properties": {n: {"type": convert_type_to_spec(p.type)} for (n, p) in body_request_params}
        }

    request_body = {}
    if schema:
        request_body["application/json"] = {"schema": schema}
        request_body["multipart/form-data"] = {"schema": schema}

    if file_request_params:
        file_properties = {}
        for pname, p in file_request_params:
            file_properties[pname] = {"type": "array", "items": {"type": "string", "format": "binary"}}
        request_body.setdefault("multipart/form-data", {}).setdefault("schema", {})\
            .setdefault("properties", {}).update(file_properties)

    o = {"operationId": endpoint,
         "parameters": params,
         "responses": responses}
    if func.__doc__:
        o['description'] = func.__doc__
    if request_body:
        o["requestBody"] = {"content": request_body}
    return o


def build_spec_param(name, request_param, loc="query"):
    o = {"name": name,
        "schema": {"type": convert_type_to_spec(request_param.type)},
        "required": loc == "path" or bool(request_param.required),
        "in": loc}
    if request_param.help:
        o['description'] = request_param.help
    return o


_url_arg_re = re.compile(r"<([a-z]+:)?([a-z0-9_]+)>")
def convert_url_args(url):
    return _url_arg_re.sub(r"{\2}", url)


def convert_type_to_spec(type):
    if type is int:
        return "integer"
    if type is float:
        return "number"
    if type is bool:
        return "boolean"
    return "string"



from frasco.ext import *
from frasco.utils import import_string
from frasco.models import delayed_tx_calls
from flask_rq2 import RQ
from flask_rq2 import cli
from rq import get_current_job
from rq.timeouts import JobTimeoutException
from .job import FrascoJob
import redis.exceptions
import functools
import logging


logger = logging.getLogger('frasco.tasks')


class FrascoTasks(Extension):
    name = 'frasco_tasks'
    prefix_extra_options = 'RQ_'
    defaults = {"tasks_timeout": RQ.default_timeout,
                "scheduled_tasks_timeout": 300}

    def _init_app(self, app, state):
        if not app.config.get('RQ_REDIS_URL') and has_extension('frasco_redis', app):
            app.config['RQ_REDIS_URL'] = app.extensions.frasco_redis.options['url']

        app.config.setdefault('RQ_JOB_CLASS', 'frasco.tasks.job.FrascoJob')
        if app.testing:
            app.config.setdefault('RQ_ASYNC', False)
        state.rq = RQ(app, default_timeout=state.options['tasks_timeout'])
        state.rq.exception_handler(per_task_exception_handler)


def per_task_exception_handler(job, *exc_info):
    if hasattr(job.func, '__task_exception_handler__'):
        return job.func.__task_exception_handler__(job, *exc_info)


def enqueue_task_now(func, *args, **kwargs):
    return enqueue_now(func, args=args, kwargs=kwargs)


def enqueue_now(func, **options):
    if getattr(func, '__task_options__', None):
        options.update(func.__task_options__)
    queue_name = options.pop('queue', None)
    if callable(queue_name):
        queue_name = queue_name()
    return get_extension_state('frasco_tasks').rq.get_queue(queue_name).enqueue_call(func, **options)


def enqueue_task(func, *args, **kwargs):
    return enqueue(func, args=args, kwargs=kwargs)


@delayed_tx_calls.proxy
def enqueue(func, **options):
    return enqueue_now(func, **options)


def get_enqueued_job(id):
    state = get_extension_state('frasco_tasks')
    return FrascoJob.fetch(id, connection=state.rq.connection)


def task(**options):
    def wrapper(func):
        func.__task_options__ = options
        setattr(func, 'enqueue', functools.partial(enqueue_task, func))
        setattr(func, 'enqueue_now', functools.partial(enqueue_task_now, func))
        setattr(func, 'exception_handler', task_exception_handler(func))
        setattr(func, 'timeout_handler', task_exception_handler(func, JobTimeoutException))
        return func
    return wrapper


def task_exception_handler(task_func, exc_type=None):
    def decorator(handler_func):
        task_func.__task_exception_handler__ = _wrap_exc_handler_for_type(handler_func, exc_type) if exc_type else handler_func
        return handler_func
    return decorator


def _wrap_exc_handler_for_type(handler_func, exc_type):
    @functools.wraps(handler_func)
    def wrapper(job, *exc_info):
        if exc_info[0] is exc_type:
            return handler_func(job, *exc_info)
    return wrapper


def schedule_task(pattern, import_name_or_func):
    state = get_extension_state('frasco_tasks')
    if isinstance(import_name_or_func, str):
        func = import_string(import_name_or_func)
        import_name = import_name_or_func
    else:
        func = import_name_or_func
        import_name = "%s.%s" % (func.__module__, func.__name__)
    try:
        logger.info("Scheduling task %s at %s" % (import_name, pattern))
        return state.rq.get_scheduler().cron(pattern, func,
            id="cron-%s" % import_name.replace('.', '-'),
            timeout=state.options['scheduled_tasks_timeout'])
    except redis.exceptions.ConnectionError:
        logger.error("Cannot initialize scheduled tasks as no redis connection is available")


def clear_all_scheduled_tasks():
    scheduler = get_extension_state('frasco_tasks').rq.get_scheduler()
    for job in scheduler.get_jobs():
        scheduler.cancel(job)


_rq2_scheduler = cli.scheduler

@functools.wraps(_rq2_scheduler)
def _scheduler(*args, **kwargs):
    state = get_extension_state('frasco_tasks')
    clear_all_scheduled_tasks()
    for import_name, pattern in state.options.get('schedule', {}).items():
        schedule_task(pattern, import_name)
    _rq2_scheduler(*args, **kwargs)

cli._commands['scheduler'] = _scheduler



from frasco.ext import get_extension_state
from frasco.models import as_transaction
from frasco.billing.invoicing import create_invoice
from frasco.billing.eu_vat import is_eu_country, get_vat_rate, should_charge_vat
import datetime


def create_invoice_from_charge(charge, obj=None, lines=None, tax_rate=None, tax_amount=None):
    state = get_extension_state('frasco_stripe')
    with create_invoice(**state.options['invoice_ref_kwargs']) as (invoice, InvoiceItem):
        invoice.currency = charge.currency.upper()
        invoice.subtotal = charge.amount / 100.0
        invoice.total = charge.amount / 100.0
        invoice.description = charge.description
        invoice.issued_at = datetime.datetime.fromtimestamp(charge.created)
        invoice.charge_id = charge.id
        invoice.paid = charge.status == "succeeded"
        if obj is not None:
            _fill_invoice_from_obj(invoice, obj)

        if tax_rate == "eu_vat":
            if is_eu_country(invoice.country):
                tax_rate = get_vat_rate(invoice.country)
                if not should_charge_vat(invoice.country, obj.eu_vat_number):
                    tax_rate = None
            else:
                tax_rate = None

        if tax_amount:
            invoice.tax_amount = tax_amount
            invoice.subtotal = invoice.total - tax_amount
        elif tax_rate:
            invoice.subtotal = invoice.total * (100 / (100 + tax_rate));
            invoice.tax_amount = invoice.total - invoice.subtotal

        if lines:
            for line in lines:
                item = InvoiceItem()
                item.amount = line['amount']
                item.quantity = line.get('quantity', 1)
                item.currency = line.get('currency', charge.currency.upper())
                item.description = line['description']
                item.tax_amount = line.get('tax_amount', (line['amount'] * tax_rate / 100) if tax_rate else 0)
                item.tax_rate = line.get('tax_rate', tax_rate or 0)
                invoice.items.append(item)


def create_invoice_from_stripe(obj, stripe_invoice):
    state = get_extension_state('frasco_stripe')
    with create_invoice(**state.options['invoice_ref_kwargs']) as (invoice, InvoiceItem):
        _fill_invoice_from_obj(invoice, obj)

        invoice.external_id = stripe_invoice.id
        invoice.currency = stripe_invoice.currency.upper()
        invoice.subtotal = stripe_invoice.subtotal / 100.0
        invoice.total = stripe_invoice.total / 100.0
        invoice.tax_amount = stripe_invoice.tax / 100.0 if stripe_invoice.tax else None
        invoice.description = stripe_invoice.description
        invoice.issued_at = datetime.datetime.fromtimestamp(stripe_invoice.created)
        invoice.paid = stripe_invoice.paid
        invoice.charge_id = stripe_invoice.charge

        for line in stripe_invoice.lines.data:
            item = InvoiceItem()
            item.external_id = line.id
            item.amount = line.amount / 100.0
            item.tax_amount = 0
            item.tax_rate = 0
            item.quantity = line.quantity
            item.currency = line.currency
            item.description = line.description or ''

            if line.tax_amounts:
                item.tax_amount = line.tax_amounts[0].amount
                item.tax_rate = line.tax_rates[0].percentage

            invoice.items.append(item)

def _fill_invoice_from_obj(invoice, obj):
    invoice.customer = obj
    invoice.email = getattr(obj, obj.__stripe_email_property__)
    if getattr(obj, '__has_stripe_billing_fields__', False):
        invoice.name = obj.billing_name
        invoice.address_line1 = obj.billing_address_line1
        invoice.address_line2 = obj.billing_address_line2
        invoice.address_city = obj.billing_address_city
        invoice.address_state = obj.billing_address_state
        invoice.address_zip = obj.billing_address_zip
        invoice.address_country = obj.billing_address_country
        if obj.billing_country:
            invoice.country = obj.billing_country.upper() 
        elif obj.billing_address_country:
            invoice.country = obj.billing_address_country.upper()



from setuptools import setup, find_packages


setup(
    name='frasco',
    version='2.0.4',
    url='http://github.com/frascoweb/frasco',
    license='MIT',
    author='Maxime Bouroumeau-Fuseau',
    author_email='maxime.bouroumeau@gmail.com',
    description='Set of extensions for Flask to develop SaaS applications',
    packages=find_packages(),
    package_data={
        'frasco': [
            'angular/static/*.js',
            'assets/*.js',
            'assets/*.html',
            'billing/invoicing/emails/*',
            'billing/stripe/emails/*',
            'mail/templates/*.html',
            'mail/templates/layouts/*',
            'push/static/*.js',
            'templating/*.html',
            'templating/bootstrap/*.html',
            'users/emails/users/*',
            'users/templates/users/*.html'
        ],
    },
    zip_safe=False,
    platforms='any',
    install_requires=[
        'ago~=0.0.93',
        'apispec~=2.0.2',
        'authlib~=0.14.1',
        'boto3~=1.12.1',
        'cssmin~=0.2.0',
        'eventlet~=0.25.1',
        'Flask~=1.1.1',
        'Flask-Assets==2.0',
        'Flask-Babel~=1.0.0',
        'Flask-Bcrypt~=0.7.1',
        'Flask-CORS~=3.0.8',
        'Flask-Login~=0.5.0',
        'Flask-Mail~=0.9.1',
        'Flask-Migrate~=2.5.2',
        'Flask-RQ2~=18.3',
        'Flask-SQLAlchemy~=2.4.1',
        'Flask-WTF~=0.14.3',
        'geoip2~=3.0.0',
        'glob2~=0.7.0',
        'goslate~=1.5.1',
        'htmlmin~=0.1.12',
        'inflection~=0.3.1',
        'jinja-layout~=0.2',
        'jinja-macro-tags~=0.2',
        'jsmin~=2.2.2',
        'Markdown~=3.1.1',
        'premailer~=3.6.1',
        'psycopg2-binary~=2.8.4',
        'pyotp~=2.3.0',
        'python-dateutil~=2.8.1',
        'python-slugify~=4.0.0',
        'python-socketio~=4.6.0',
        'PyYAML~=5.3',
        'redis~=3.5.3',
        'requests~=2.22.0',
        'simplejson~=3.17.0',
        'suds-py3~=1.4.1.0',
        'speaklater~=1.3',
        'stripe~=2.42.0',
        'werkzeug>=1.0.0',
    ],
    entry_points='''
        [console_scripts]
        frasco=flask.cli:main
    '''
)
